#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>

#include "mmio_mgt.h"
#include "list.h"

#define DEV_PATH                    "/dev/mm_device"

// From mm_driver.c
#define IOCTL_RESERVE_PHYSICAL  _IO('W', 0)
#define IOCTL_RELEASE_PHYSICAL  _IO('W', 1)

#define PAGE_SHIFT              (12)
#define PAGE_SIZE               ((size_t)0x1000ull)
#define PAGE_MASK               (~(PAGE_SIZE-1))
#define ROUND_PAGE(size)        ((size + (PAGE_SIZE-1)) & PAGE_MASK)

struct RESERVE_PHYSICAL {
    union {
        size_t physical_size;     // IN
        off_t physical_address;  // OUT
    };
};

typedef struct mapped_item {
    struct list_head list;
    void *vaddr;
    off_t paddr;
    size_t size;
} mapped_item_t;

static int mm_driver_fd = -1; // Driver file descriptor

int init_mm_driver(void)
{
    mm_driver_fd = open(DEV_PATH, O_RDWR);

    if (mm_driver_fd < 0)
    {
        perror("open()");
        return -1;
    }

    return 0;
}

int fini_mm_driver(void)
{
    close(mm_driver_fd);
    mm_driver_fd = -1;
}

int map_physical(void **p_ptr, off_t paddr, size_t psize)
{
    void *ptr = mmap(
        NULL,
        psize,
        PROT_READ | PROT_WRITE,
        MAP_SHARED | MAP_FILE,
        mm_driver_fd,
        paddr);

    if (ptr == MAP_FAILED)
    {
        perror("mmap()");
        return -1;
    }

    *p_ptr = ptr;

    return 0;
}

int unmap_physical(void *ptr, size_t size)
{
    return munmap(ptr, size);
}

int map_bar0_buffer(void **p_mmio_area)
{
    return map_physical(p_mmio_area, MMIO_BUFFER_PADDR, MMIO_BUFFER_SIZE);
}

int reserve_physical(size_t size, off_t *p_paddr)
{
    struct RESERVE_PHYSICAL param = {
        .physical_size = size
    };

    if (ioctl(mm_driver_fd, IOCTL_RESERVE_PHYSICAL, &param) < 0)
    {
        perror("ioctl()");
        return -1;
    }

    *p_paddr = param.physical_address;

    return 0;
}

int release_physical(off_t paddr)
{
    if (ioctl(mm_driver_fd, IOCTL_RELEASE_PHYSICAL, &paddr) < 0)
    {
        perror("ioctl()");
        return -1;
    }

    return 0;
}

LIST_HEAD(mapped_item_list);

int allocate_and_map_physical(size_t size, mem_descriptor_t **p_memdesc)
{
    int res = 0;
    mem_descriptor_t *memdesc = calloc(1, sizeof(mem_descriptor_t));
    size_t rounded_size = ROUND_PAGE(size);
    off_t paddr = 0;
    void *vaddr = NULL;

    if (memdesc == NULL)
    {
        perror("allocate_and_map_physical calloc()");
        return -1;
    }

    if (reserve_physical(rounded_size, &paddr))
    {
        printf("Failed to reserver physical memory\n");
        free(memdesc);
        return -1;
    }

    if (map_physical(&vaddr, paddr, rounded_size) < 0)
    {
        printf("Failed to map physical off %lx\n", paddr);
        res = -1;
        goto out;
    }

    memset(vaddr, 0, rounded_size);

    /* Initialise the memory descriptor. */
    memdesc->paddr = paddr;
    memdesc->size = size;
    memdesc->rounded_size = rounded_size;
    memdesc->vaddr = vaddr;
    list_add(&memdesc->list, &mapped_item_list);

    *p_memdesc = memdesc;

out:
    if (res < 0)
    {
        release_physical(paddr);
        
        if (vaddr)
            unmap_physical(vaddr, rounded_size);
    }

    return res;
}

int deallocate_and_unmap_physical(mem_descriptor_t *memdesc)
{
    unmap_physical(memdesc->vaddr, memdesc->rounded_size);

    release_physical(memdesc->paddr);

    list_del(&memdesc->list);

    free(memdesc);

    return 0;
}